@@ -4,6 +4,7 @@
 #include "geos.h"
 
 #include <Python.h>
+#include <numpy/ndarraytypes.h>
 #include <numpy/npy_math.h>
 #include <structmember.h>
 
@@ -25,6 +26,12 @@ void destroy_geom_arr(void* context, GEOSGeometry** array, int length) {
   }
 }
 
+/* These functions are used to workaround two GEOS issues (in WKB writer for
+ * GEOS < 3.9, in WKT writer for GEOS < 3.9 and in GeoJSON writer for GEOS 3.10.0):
+ * - POINT EMPTY was not handled correctly (we do it ourselves)
+ * - MULTIPOINT (EMPTY) resulted in segfault (we check for it and raise)
+ */
+
 /* Returns 1 if a multipoint has an empty point, 0 otherwise, 2 on error.
  */
 char multipoint_has_point_empty(GEOSContextHandle_t ctx, GEOSGeometry* geom) {
@@ -50,10 +57,6 @@ char multipoint_has_point_empty(GEOSContextHandle_t ctx, GEOSGeometry* geom) {
   return 0;
 }
 
-// POINT EMPTY is converted to POINT (nan nan)
-// by GEOS >= 3.10.0. Before that, we do it ourselves here.
-#if !GEOS_SINCE_3_10_0
-
 /* Returns 1 if geometry is an empty point, 0 otherwise, 2 on error.
  */
 char is_point_empty(GEOSContextHandle_t ctx, GEOSGeometry* geom) {
@@ -255,8 +258,6 @@ GEOSGeometry* point_empty_to_nan_all_geoms(GEOSContextHandle_t ctx, GEOSGeometry
   return result;
 }
 
-#endif  // !GEOS_SINCE_3_10_0
-
 /* Checks whether the geometry is a multipoint with an empty point in it
  *
  * According to https://github.com/libgeos/geos/issues/305, this check is not
@@ -289,6 +290,74 @@ char check_to_wkt_compatible(GEOSContextHandle_t ctx, GEOSGeometry* geom) {
   }
 }
 
+#if GEOS_SINCE_3_9_0
+
+/* Checks whether the geometry is a 3D empty geometry and, if so, create the WKT string
+ *
+ * GEOS 3.9.* is able to distiguish 2D and 3D simple geometries (non-collections). But the
+ * but the WKT serialization never writes a 3D empty geometry. This function fixes that.
+ * It only makes sense to use this for GEOS versions >= 3.9.
+ *
+ * Pending GEOS ticket: https://trac.osgeo.org/geos/ticket/1129
+ *
+ * If the geometry is a 3D empty, then the (char**) wkt argument is filled with the
+ * correct WKT string. Else, wkt becomes NULL and the GEOS WKT writer should be used.
+ *
+ * The return value is one of:
+ * - PGERR_SUCCESS
+ * - PGERR_GEOS_EXCEPTION
+ */
+char wkt_empty_3d_geometry(GEOSContextHandle_t ctx, GEOSGeometry* geom, char** wkt) {
+  char is_empty;
+  int geom_type;
+
+  is_empty = GEOSisEmpty_r(ctx, geom);
+  if (is_empty == 2) {
+    return PGERR_GEOS_EXCEPTION;
+  } else if (is_empty == 0) {
+    *wkt = NULL;
+    return PGERR_SUCCESS;
+  }
+  if (GEOSGeom_getCoordinateDimension_r(ctx, geom) == 2) {
+    *wkt = NULL;
+    return PGERR_SUCCESS;
+  }
+  geom_type = GEOSGeomTypeId_r(ctx, geom);
+  switch (geom_type) {
+    case GEOS_POINT:
+      *wkt = "POINT Z EMPTY";
+      return PGERR_SUCCESS;
+    case GEOS_LINESTRING:
+      *wkt = "LINESTRING Z EMPTY";
+      break;
+    case GEOS_LINEARRING:
+      *wkt = "LINEARRING Z EMPTY";
+      break;
+    case GEOS_POLYGON:
+      *wkt = "POLYGON Z EMPTY";
+      break;
+    // Note: Empty collections cannot be 3D in GEOS.
+    // We do include the options in case of future support.
+    case GEOS_MULTIPOINT:
+      *wkt = "MULTIPOINT Z EMPTY";
+      break;
+    case GEOS_MULTILINESTRING:
+      *wkt = "MULTILINESTRING Z EMPTY";
+      break;
+    case GEOS_MULTIPOLYGON:
+      *wkt = "MULTIPOLYGON Z EMPTY";
+      break;
+    case GEOS_GEOMETRYCOLLECTION:
+      *wkt = "GEOMETRYCOLLECTION Z EMPTY";
+      break;
+    default:
+      return PGERR_GEOS_EXCEPTION;
+  }
+  return PGERR_SUCCESS;
+}
+
+#endif  // GEOS_SINCE_3_9_0
+
 /* GEOSInterpolate_r and GEOSInterpolateNormalized_r segfault on empty
  * geometries and also on collections with the first geometry empty.
  *
@@ -310,8 +379,8 @@ char geos_interpolate_checker(GEOSContextHandle_t ctx, GEOSGeometry* geom) {
   type = GEOSGeomTypeId_r(ctx, geom);
   if (type == -1) {
     return PGERR_GEOS_EXCEPTION;
-  } else if ((type == GEOS_POINT) || (type == GEOS_POLYGON) || (type == GEOS_MULTIPOINT) ||
-             (type == GEOS_MULTIPOLYGON)) {
+  } else if ((type == GEOS_POINT) || (type == GEOS_POLYGON) ||
+             (type == GEOS_MULTIPOINT) || (type == GEOS_MULTIPOLYGON)) {
     return PGERR_GEOMETRY_TYPE;
   }
 
@@ -595,3 +664,277 @@ GEOSGeometry* create_point(GEOSContextHandle_t ctx, double x, double y) {
   return geom;
 #endif
 }
+
+/* Create a 3D empty Point
+ *
+ * Works around a limitation of the GEOS C API by constructing the point
+ * from its WKT representation (POINT Z EMPTY).
+ *
+ * Returns
+ * -------
+ * GEOSGeometry* on success (owned by caller) or NULL on failure
+ */
+GEOSGeometry* PyGEOS_create3DEmptyPoint(GEOSContextHandle_t ctx) {
+  const char* wkt = "POINT Z EMPTY";
+  GEOSGeometry* geom;
+  GEOSWKTReader* reader;
+
+  reader = GEOSWKTReader_create_r(ctx);
+  if (reader == NULL) {
+    return NULL;
+  }
+  geom = GEOSWKTReader_read_r(ctx, reader, wkt);
+  GEOSWKTReader_destroy_r(ctx, reader);
+  return geom;
+}
+
+/* Force the coordinate dimensionality (2D / 3D) of any geometry
+ *
+ * Parameters
+ * ----------
+ * ctx: GEOS context handle
+ * geom: geometry
+ * dims: dimensions to force (2 or 3)
+ * z: Z coordinate (ignored if dims==2)
+ *
+ * Returns
+ * -------
+ * GEOSGeometry* on success (owned by caller) or NULL on failure
+ */
+GEOSGeometry* force_dims(GEOSContextHandle_t, GEOSGeometry*, unsigned int, double);
+
+GEOSGeometry* force_dims_simple(GEOSContextHandle_t ctx, GEOSGeometry* geom, int type,
+                                unsigned int dims, double z) {
+  unsigned int actual_dims, n, i, j;
+  double coord;
+  const GEOSCoordSequence* seq = GEOSGeom_getCoordSeq_r(ctx, geom);
+
+  /* Special case for POINT EMPTY (Point coordinate list cannot be 0-length) */
+  if ((type == 0) && (GEOSisEmpty_r(ctx, geom) == 1)) {
+    if (dims == 2) {
+      return GEOSGeom_createEmptyPoint_r(ctx);
+    } else if (dims == 3) {
+      return PyGEOS_create3DEmptyPoint(ctx);
+    } else {
+      return NULL;
+    }
+  }
+
+  /* Investigate the coordinate sequence, return when already of correct dimensionality */
+  if (GEOSCoordSeq_getDimensions_r(ctx, seq, &actual_dims) == 0) {
+    return NULL;
+  }
+  if (actual_dims == dims) {
+    return GEOSGeom_clone_r(ctx, geom);
+  }
+  if (GEOSCoordSeq_getSize_r(ctx, seq, &n) == 0) {
+    return NULL;
+  }
+
+  /* Create a new one to fill with the new coordinates */
+  GEOSCoordSequence* seq_new = GEOSCoordSeq_create_r(ctx, n, dims);
+  if (seq_new == NULL) {
+    return NULL;
+  }
+
+  for (i = 0; i < n; i++) {
+    for (j = 0; j < 2; j++) {
+      if (!GEOSCoordSeq_getOrdinate_r(ctx, seq, i, j, &coord)) {
+        GEOSCoordSeq_destroy_r(ctx, seq_new);
+        return NULL;
+      }
+      if (!GEOSCoordSeq_setOrdinate_r(ctx, seq_new, i, j, coord)) {
+        GEOSCoordSeq_destroy_r(ctx, seq_new);
+        return NULL;
+      }
+    }
+    if (dims == 3) {
+      if (!GEOSCoordSeq_setZ_r(ctx, seq_new, i, z)) {
+        GEOSCoordSeq_destroy_r(ctx, seq_new);
+        return NULL;
+      }
+    }
+  }
+
+  /* Construct a new geometry */
+  if (type == 0) {
+    return GEOSGeom_createPoint_r(ctx, seq_new);
+  } else if (type == 1) {
+    return GEOSGeom_createLineString_r(ctx, seq_new);
+  } else if (type == 2) {
+    return GEOSGeom_createLinearRing_r(ctx, seq_new);
+  } else {
+    return NULL;
+  }
+}
+
+GEOSGeometry* force_dims_polygon(GEOSContextHandle_t ctx, GEOSGeometry* geom,
+                                 unsigned int dims, double z) {
+  int i, n;
+  const GEOSGeometry *shell, *hole;
+  GEOSGeometry *new_shell, *new_hole, *result = NULL;
+  GEOSGeometry** new_holes;
+
+  n = GEOSGetNumInteriorRings_r(ctx, geom);
+  if (n == -1) {
+    return NULL;
+  }
+
+  /* create the exterior ring */
+  shell = GEOSGetExteriorRing_r(ctx, geom);
+  if (shell == NULL) {
+    return NULL;
+  }
+  new_shell = force_dims_simple(ctx, (GEOSGeometry*)shell, 2, dims, z);
+  if (new_shell == NULL) {
+    return NULL;
+  }
+
+  new_holes = malloc(sizeof(void*) * n);
+  if (new_holes == NULL) {
+    GEOSGeom_destroy_r(ctx, new_shell);
+    return NULL;
+  }
+
+  for (i = 0; i < n; i++) {
+    hole = GEOSGetInteriorRingN_r(ctx, geom, i);
+    if (hole == NULL) {
+      GEOSGeom_destroy_r(ctx, new_shell);
+      destroy_geom_arr(ctx, new_holes, i - 1);
+      goto finish;
+    }
+    new_hole = force_dims_simple(ctx, (GEOSGeometry*)hole, 2, dims, z);
+    if (hole == NULL) {
+      GEOSGeom_destroy_r(ctx, new_shell);
+      destroy_geom_arr(ctx, new_holes, i - 1);
+      goto finish;
+    }
+    new_holes[i] = new_hole;
+  }
+
+  result = GEOSGeom_createPolygon_r(ctx, new_shell, new_holes, n);
+
+finish:
+  if (new_holes != NULL) {
+    free(new_holes);
+  }
+  return result;
+}
+
+GEOSGeometry* force_dims_collection(GEOSContextHandle_t ctx, GEOSGeometry* geom, int type,
+                                    unsigned int dims, double z) {
+  int i, n;
+  const GEOSGeometry* sub_geom;
+  GEOSGeometry *new_sub_geom, *result = NULL;
+  GEOSGeometry** geoms;
+
+  n = GEOSGetNumGeometries_r(ctx, geom);
+  if (n == -1) {
+    return NULL;
+  }
+
+  geoms = malloc(sizeof(void*) * n);
+  if (geoms == NULL) {
+    return NULL;
+  }
+
+  for (i = 0; i < n; i++) {
+    sub_geom = GEOSGetGeometryN_r(ctx, geom, i);
+    if (sub_geom == NULL) {
+      destroy_geom_arr(ctx, geoms, i - i);
+      goto finish;
+    }
+    new_sub_geom = force_dims(ctx, (GEOSGeometry*)sub_geom, dims, z);
+    if (new_sub_geom == NULL) {
+      destroy_geom_arr(ctx, geoms, i - i);
+      goto finish;
+    }
+    geoms[i] = new_sub_geom;
+  }
+
+  result = GEOSGeom_createCollection_r(ctx, type, geoms, n);
+finish:
+  if (geoms != NULL) {
+    free(geoms);
+  }
+  return result;
+}
+
+GEOSGeometry* force_dims(GEOSContextHandle_t ctx, GEOSGeometry* geom, unsigned int dims,
+                         double z) {
+  int type = GEOSGeomTypeId_r(ctx, geom);
+  if ((type == 0) || (type == 1) || (type == 2)) {
+    return force_dims_simple(ctx, geom, type, dims, z);
+  } else if (type == 3) {
+    return force_dims_polygon(ctx, geom, dims, z);
+  } else if ((type >= 4) && (type <= 7)) {
+    return force_dims_collection(ctx, geom, type, dims, z);
+  } else {
+    return NULL;
+  }
+}
+
+GEOSGeometry* PyGEOSForce2D(GEOSContextHandle_t ctx, GEOSGeometry* geom) {
+  return force_dims(ctx, geom, 2, 0.0);
+}
+
+GEOSGeometry* PyGEOSForce3D(GEOSContextHandle_t ctx, GEOSGeometry* geom, double z) {
+  return force_dims(ctx, geom, 3, z);
+}
+
+GEOSCoordSequence* coordseq_from_buffer(GEOSContextHandle_t ctx, const double* buf,
+                                        unsigned int size, unsigned int dims, char ring_closure,
+                                        npy_intp cs1, npy_intp cs2) {
+  GEOSCoordSequence* coord_seq;
+  char *cp1, *cp2;
+  unsigned int i, j;
+  double first_coord;
+
+#if GEOS_SINCE_3_10_0
+
+  if (!ring_closure) {
+    if ((cs1 == dims * 8) && (cs2 == 8)) {
+      /* C-contiguous memory */
+      int hasZ = dims == 3;
+      coord_seq = GEOSCoordSeq_copyFromBuffer_r(ctx, buf, size, hasZ, 0);
+      return coord_seq;
+    }
+    else if ((cs1 == 8) && (cs2 == size * 8)) {
+      /* F-contiguous memory (note: this for the subset, so we don't necessarily
+      end up here if the full array is F-contiguous) */
+      const double* x = buf;
+      const double* y =  (double*)((char*)buf + cs2);
+      const double* z = (dims == 3) ? (double*)((char*)buf + 2 * cs2) : NULL;
+      coord_seq = GEOSCoordSeq_copyFromArrays_r(ctx, x, y, z, NULL, size);
+      return coord_seq;
+    }
+  }
+
+#endif
+
+  coord_seq = GEOSCoordSeq_create_r(ctx, size + ring_closure, dims);
+  if (coord_seq == NULL) {
+    return NULL;
+  }
+  cp1 = (char*)buf;
+  for (i = 0; i < size; i++, cp1 += cs1) {
+    cp2 = cp1;
+    for (j = 0; j < dims; j++, cp2 += cs2) {
+      if (!GEOSCoordSeq_setOrdinate_r(ctx, coord_seq, i, j, *(double*)cp2)) {
+        GEOSCoordSeq_destroy_r(ctx, coord_seq);
+        return NULL;
+      }
+    }
+  }
+  /* add the closing coordinate if necessary */
+  if (ring_closure) {
+    for (j = 0; j < dims; j++){
+      first_coord = *(double*)((char*)buf + j * cs2);
+      if (!GEOSCoordSeq_setOrdinate_r(ctx, coord_seq, size, j, first_coord)) {
+        GEOSCoordSeq_destroy_r(ctx, coord_seq);
+        return NULL;
+      }
+    }
+  }
+  return coord_seq;
+}
